#!/usr/bin/python

# Copyright 2007 Yannick Gingras <ygingras@ygingras.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
# MA 02110-1301 USA

"""probabilities, frequencies, loading, and training stuff"""

FORMAT_VERSION = 2

from cPickle import load, dump
from pprint import pformat, pprint
from pkg_resources import resource_filename, resource_exists, resource_listdir
from random import random
import re
import os
import trainsets


WORD_PAT = re.compile(r"\b(\w+)\b", re.UNICODE) 

# The main data structure is a probability table.  For training, we
# also need a occurence count table.  IO is done with pickle.

class TrainSet:
    """Character based markov chain with a memory of 2 past steps."""
    
    def __init__(self, counts={}, probs={}, words=set()):
        # ex: counts[('l', 'a')] = (12, {'a':8, 'c':4})
        self.counts = counts
        # ex: probs[('c', 'r')] = [(.1, 'e'), (.07, 'a'),...,(.001, 'z')]
        self.probs  = probs
        self.words  = words

    def save(self, path):
        dump(dict(version=FORMAT_VERSION,
                  counts=self.counts,
                  probs=self.probs,
                  words=self.words),
             open(path, "wb"),
             -1)

    def learn(self, data, count_only=False, min_len=3, max_len=14):
        """Add the transition frequencies of data to the prob tables.

        Data should be human readable unicode text.
        """
        # count
        for match in WORD_PAT.finditer(data):
            word = match.group()
            # trim digits (too ugly with the regexp)
            word = u"".join(filter(lambda c:c.isalpha(), word))
            if not (min_len <= len(word) <= max_len):
                continue

            self.words.add(word.lower())
            
            # the first two letters of a word also have transitions
            prev = pprev = None
            for c in word:
                nb_trs, trs_h  = self.counts.get((pprev, prev), (0, {}))
                trs_h[c] = trs_h.get(c, 0) + 1
                self.counts[(pprev, prev)] = (nb_trs+1, trs_h)

                pprev = prev
                prev = c

            # transition to word end
            nb_trs, trs_h  = self.counts.get((pprev, prev), (0, {}))
            trs_h[None] = trs_h.get(None, 0) + 1
            self.counts[(pprev, prev)] = (nb_trs+1, trs_h)
            
        if not count_only:
            self.update_probs()

    def update_probs(self):
        self.probs = {}
        for k in self.counts.keys():
            nb_trs, trs_h = self.counts[k]
            self.probs[k] = [ (1.0*freq/nb_trs, nc)
                              for nc, freq in trs_h.items()]
            self.probs[k].sort()
            self.probs[k].reverse()


    def get_c(self, trans, draw):
        tot = 0
        for prob, c in trans:
            tot += prob
            if tot >= draw:
                return c
        return c


    def gen(self, non_word=True, minl=3, maxl=18):
        choices = []
        pprev = prev = None
        c = self.get_c(self.probs[(None, None)], random())
        while c != None:
            choices.append(c)
            pprev = prev
            prev = c
            c = self.get_c(self.probs[(pprev, prev)], random())

        word = "".join(choices)
        if non_word and word.lower() in self.words or not(minl <= len(word) <= maxl):
            return self.gen(non_word, minl, maxl)
        return word


def list_trainsets():
    return [f for f, e in
            filter(lambda t:t[1] == ".yould",
                   map(os.path.splitext,
                       resource_listdir(trainsets.__name__, ".")))]
    
            
def find_trainset(name):
    if os.path.isfile(name):
        return load_trainset(name)
    elif resource_exists(trainsets.__name__, name):
        return load_trainset(resource_filename(trainsets.__name__, name))
    elif resource_exists(trainsets.__name__, name+".yould"):
        return load_trainset(resource_filename(trainsets.__name__, name+".yould"))
    else:
        raise Exception("Can't find training set '%s'" % name)
    

def load_trainset(path):
    data = load(open(path, "rb"))
    if data["version"] != FORMAT_VERSION:
        raise Exception("This file format is not supported.")
    return TrainSet(data["counts"], data["probs"], data["words"])




#!/usr/bin/python

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
# MA 02110-1301 USA

from urllib import urlopen
from re import compile

# King James Bible
BOOKS = [8010, 8011, 8012, 8013, 8014, 8015, 8016, 8017, 8018, 8019,
         8001, 8020, 8021, 8022, 8023, 8024, 8025, 8026, 8027, 8028,
         8029, 8002, 8030, 8031, 8032, 8033, 8034, 8035, 8036, 8037,
         8038, 8039, 8003, 8040, 8041, 8042, 8043, 8044, 8045, 8046,
         8047, 8048, 8049, 8004, 8050, 8051, 8052, 8053, 8054, 8055,
         8056, 8057, 8058, 8059, 8005, 8060, 8061, 8062, 8063, 8064,
         8065, 8066, 8006, 8007, 8008, 8009]
  

BASE_URL = "http://www.gutenberg.org/"
MAIN_URL = BASE_URL + "etext/%d"
DL_PAT = compile(r'<td class="pgdbfilesdownload"><a href="(/(?:dirs|files)/.*?.txt)"')
MAX = 2**22

for book in BOOKS:
    try:
        data = urlopen(MAIN_URL % book).read(MAX)
        print book, 
        dl = DL_PAT.findall(data)[0]
        print dl
        stream = urlopen(BASE_URL + dl)
        open("%04d.txt" % book, "w").write(stream.read(MAX))
    except:
        print "Error"
    



#!/usr/bin/python

# Copyright 2007 Yannick Gingras <ygingras@ygingras.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
# MA 02110-1301 USA

"""utility code to check if a domain is available"""

# naming convention used:
#  - name: a word without the tdl; ex: ygingras
#  - tld:  top level domain without the "."; ex: net
#  - dom:  a domain name with the tld; ex: ygingras.net

import os
from time import sleep
from os import popen
from threading import Thread
import DNS

TIMEOUT = 1.5
WHOIS = "/usr/bin/whois %s"
HOST_SOA = "/usr/bin/host -C %s 2>&1"
#HOST_SOA = "host -C %s "
WHOIS_ERR = dict(com="No match for",
                 net="No match for",                 
                 org="NOT FOUND",
                 info="NOT FOUND",
                 ca="Status:         AVAIL",)

# TODO: should parse resolv.conf
DNS_SERVER = "192.168.0.1"

def dom_tld(dom):
    return os.path.splitext(dom)[1][1:]


def dom_name(dom):
    return os.path.splitext(dom)[0]


def encode(dom):
    try:
        asc = dom.encode("ascii")
        if asc == dom:
            return asc
    except UnicodeEncodeError:
        pass
    name = dom_name(dom).encode("punycode")
    return "xn--%s.%s" % (name, dom_tld(dom))


def new_has_soa(dom):
    try:
        d = DNS.DnsRequest(server=DNS_SERVER, timeout=TIMEOUT)
        r = d.req(name=encode(dom), qtype="SOA")
        return bool(r.answers)
    except DNS.Base.DNSError:
        return False


def has_soa(dom):
    """Look for a soa record.

    Most domains without a soa a available and checking for a soa is
    faster than hitting the whois database."""
    pipe = popen(HOST_SOA % encode(dom))

    res = []
    
    def check():
        rec = pipe.read()
        ecode = pipe.close()
        res.append((rec, ecode))

    th = Thread(target=check)
    th.setDaemon(True)

    th.start()
    th.join(TIMEOUT)
    if res:
        #print "* ok"
        rec, ecode = res.pop()
        return len(rec) == 0 and ecode is None
    else:
        #print "*** timeout"
        return True
    

def has_whois(dom):
    # we only check the first few characters returned by whois but
    # what we are looking for depend on the TLD.
    tld = dom_tld(dom)

    
    #pipe = popen(WHOIS % encode(dom))
    #print WHOIS % dom.encode("utf-8")
    pipe = popen(WHOIS % encode(dom))
    lines = [pipe.readline() for i in range(9)]
    #print "-" * 30
    #print "".join(lines)

    avail = len(filter(lambda l:l.startswith(WHOIS_ERR[tld]),
                       lines)) != 0

    if pipe.close():
        raise Exception("whois error")

    return not avail


def available(dom):
    return not new_has_soa(dom) and not has_whois(dom)




#!/usr/bin/python

# Copyright 2007 Yannick Gingras <ygingras@ygingras.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
# MA 02110-1301 USA

import sys

from dom import available
from prob import find_trainset, list_trainsets
from pprint import pprint
from optparse import OptionParser

def main():
    parser = OptionParser()
    
    parser.add_option("-t", "--training-set", dest="trainset", default="kjb",
                      help="encoding of the text, default is UTF-8")
    parser.add_option("-o", "--output", dest="output", default=None,
                      help="save output in FILE", metavar="FILE")
    parser.add_option("-n", dest="nb_words", default=1, type=int,
                      help="number of words to generate")
    parser.add_option("-m", "--minimum", dest="min", default=4, type=int,
                      help="minimum word length")
    parser.add_option("-M", "--maximum", dest="max", default=12, type=int,
                      help="maximum word length")
    parser.add_option("-d", "--dump",
                      action="store_true", dest="dump", default=False,
                      help="dump the training set in human readable format and exit")
    parser.add_option("-r", "--real-words",
                      action="store_true", dest="real_words", default=False,
                      help="include words seen during training")
    parser.add_option("--list-sets",
                      action="store_true", dest="list", default=False,
                      help="list the available training sets")

    (opts, args) = parser.parse_args()
    if args:
        parser.error("Unrecognized arguments: %s" % ", ".join(args))
    
    
    trainset = find_trainset(opts.trainset)

    if opts.list:
        print ", ".join(list_trainsets())
        sys.exit(0)
    
    if opts.dump:
        # TODO: we could have something more readable than that
        pprint(trainset.probs)
        sys.exit(0)

    if opts.output:
        out = open(out, "w")
    else:
        out = sys.stdout

    for i in xrange(opts.nb_words):
        word = trainset.gen(not opts.real_words, opts.min, opts.max)

        # TODO: locale dependant encoding
        out.write(word.encode("utf-8"))
        out.write("\n")

        #print ["not", "is "][available(word+".org")], "avaiable"


if __name__ == "__main__":
    main()




#!/usr/bin/python

# Copyright 2007 Yannick Gingras <ygingras@ygingras.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
# MA 02110-1301 USA

import os
import sys
from optparse import OptionParser

from prob import *

BODY_TAG = " OF THE PROJECT GUTENBERG EBOOK"

def gut_strip(file):
    """Strip project gutenberg prolog.  Generator version"""
    in_body = False
    for line in file:
        if not in_body:
            if line.find("START"+BODY_TAG) == -1:
                continue
            else:
                in_body = True
        elif line.find("END"+BODY_TAG) != -1:
            break
        else:
            yield(line)

def main():
    parser = OptionParser(usage="%prog PROB_FILE TEXT1 [TEXT2 ...]")
    
    parser.add_option("-e", "--encoding", dest="encoding", default="utf-8",
                      help="encoding of the text, default is UTF-8")
    parser.add_option("-g", "--gutenberg",
                      action="store_true", dest="gutenberg", default=False,
                      help=("assume that texts are Project Gutenberg ebooks and"
                            " skip the prolog "))
    
    (opts, args) = parser.parse_args()
    if not args:
        parser.error("Insufficient number of arguments")

    probpath = args[0]
    if os.path.isfile(probpath):
        trainset = load_trainset(probpath)
    else:
        trainset = TrainSet()
        
    for f in args[1:]:
        
        in_body = False
        file = f == "-" and sys.stdin or open(f)
        if opts.gutenberg:
            file = gut_strip(file)
        for line in file:
            line = unicode(line, opts.encoding)
            trainset.learn(line, True)

    trainset.update_probs()
    trainset.save(probpath)

if __name__ == "__main__":
    main()




#!/usr/bin/python

# Copyright 2007 Yannick Gingras <ygingras@ygingras.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
# MA 02110-1301 USA

from urllib import urlopen
from re import compile
from random import sample
from optparse import OptionParser
import sys

BASE_URL = "http://www.gutenberg.org/"
MAIN_URL = BASE_URL + "etext/%s"
LANG_URL="http://www.gutenberg.org/browse/languages/%s"
#DL_PAT = compile(r'<td class="pgdbfilesdownload"><a href="(/(?:dirs|files)/.*?.txt)"')
DL_PAT = compile(r'class="pgdbfilesencoding">(?:iso-8859-1|us-ascii)</td>.*?<td class="pgdbfilesdownload"><a href="(/(?:dirs|files)/.*?.txt)"')
LANG_PAT = compile(r'<li class="pgdbetext"><a href="/etext/(.+?)">')
MAX = 2**22

def fetch(url):
    return urlopen(url).read(MAX)

def getsample(books, nb_books):
    for book in sample(books, min(len(books), nb_books)):
        print book, 
        try:
            data = fetch(MAIN_URL % book)
            dl = DL_PAT.findall(data)[0]
            open("%s.txt" % book, "w").write(fetch(BASE_URL + dl))
            print BASE_URL + dl
        except:
            print "* Error *"

# Langs with enough stuff: zh, nl, en, fi, fr, de, it, pt, es, tl
# Langs that won't parse: zh, en
def getlang(lang):
    return LANG_PAT.findall(fetch(LANG_URL % lang))

def main():
    parser = OptionParser(usage="%prog LANG")
    
    parser.add_option("-n", "--sample-size", dest="nb_books", default=50,
                      type=int,
                      help="number of books to download, default is 50")
    
    (opts, args) = parser.parse_args()
    if len(args) != 1:
        parser.print_usage()
        parser.exit()
    
    getsample(getlang(args[0]), opts.nb_books)

if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright 2007 Yannick Gingras <ygingras@ygingras.net>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
# MA 02110-1301 USA

from ez_setup import use_setuptools
use_setuptools()
from setuptools import setup, find_packages

setup(
    name='yould',
    version="0.3.1",
    description="a word generator",
    author="Yannick Gingras",
    author_email="ygingras@ygingras.net",
    url="http://ygingras.net/b/tag/yould",
    license='GPL v2 or later',
    install_requires=[],
    packages=find_packages(),
    include_package_data=True,
    package_data = {
       '': ['*.yould'],
    },
    entry_points={
      'console_scripts': ["yould = yould.generate:main",
                          "yould-train = yould.train:main"],
    },
)
